from __future__ import absolute_import
import json

from jet_bridge_base.fields.field import Field


class ArrayField(Field):
    field_error_messages = {
        'invalid': 'not a valid array'
    }

    def to_internal_value_item(self, value):
        try:
            result = json.loads(value)
            if not isinstance(result, list):
                raise ValueError
            return result
        except ValueError:
            self.error('invalid')

    def to_representation_item(self, value):
        return json.dumps(value)



from sqlalchemy import inspect, desc
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.sql import operators, text
from sqlalchemy.sql.elements import AnnotatedColumnElement, UnaryExpression
import dateparser
from datetime import datetime

from jet_bridge_base import settings


def get_queryset_order_by(queryset):
    if hasattr(queryset, '_order_by_clauses') and queryset._order_by_clauses:
        return queryset._order_by_clauses
    elif hasattr(queryset, '_order_by') and queryset._order_by:
        return queryset._order_by
    else:
        return []


def get_queryset_limit(queryset):
    if hasattr(queryset, '_limit_clause') and queryset._limit_clause:
        return queryset._limit_clause
    elif hasattr(queryset, '_limit') and queryset._limit:
        return queryset._limit


def apply_default_ordering(Model, queryset):
    mapper = inspect(Model)
    pk = mapper.primary_key[0]
    ordering = get_queryset_order_by(queryset)

    def is_pk(x):
        if isinstance(x, AnnotatedColumnElement):
            return x.name == pk.name
        elif isinstance(x, UnaryExpression):
            return hasattr(x.element, 'name') and x.element.name == pk.name and x.modifier == operators.desc_op
        return False

    if ordering is None or not any(map(is_pk, ordering)):
        queryset = queryset.order_by(desc(pk))

    return queryset


def queryset_count_optimized_for_postgresql(request, db_table):
    try:
        cursor = request.session.execute(text('SELECT reltuples FROM pg_class WHERE relname = :db_table'), {'db_table': db_table})
        row = cursor.fetchone()
        return int(row[0])
    except SQLAlchemyError:
        request.session.rollback()
        raise


def queryset_count_optimized_for_mysql(request, db_table):
    try:
        cursor = request.session.execute(text('EXPLAIN SELECT COUNT(*) FROM `{}`'.format(db_table)))
        row = cursor.fetchone()
        return int(row[8])
    except SQLAlchemyError:
        request.session.rollback()
        raise


def queryset_count_optimized(request, queryset):
    if queryset.whereclause is None:
        count_optimized = None

        try:
            table = queryset.statement.froms[0].name

            if get_session_engine(queryset.session) == 'postgresql':
                count_optimized = queryset_count_optimized_for_postgresql(request, table)
            elif get_session_engine(queryset.session) == 'mysql':
                count_optimized = queryset_count_optimized_for_mysql(request, table)
        except:
            pass

        if count_optimized is not None and count_optimized >= 10000:
            return count_optimized

    try:
        return queryset.count()
    except SQLAlchemyError:
        queryset.session.rollback()
        raise


def get_session_engine(session):
    return session.bind.engine.name


def apply_session_timezone(session, timezone):
    if get_session_engine(session) == 'mysql':
        session.execute('SET time_zone = :tz', {'tz': timezone})
    elif get_session_engine(session) in ['postgresql', 'mssql']:
        offset_hours = dateparser.parse(datetime.now().isoformat() + timezone).utcoffset().total_seconds() / 60 / 60
        offset_hours_str = '{:+}'.format(offset_hours).replace(".0", "")
        session.execute('SET TIME ZONE :tz', {'tz': offset_hours_str})



from jet_bridge_base.exceptions.validation_error import ValidationError
from sqlalchemy import func, sql, inspect
from sqlalchemy.sql import sqltypes, text

from jet_bridge_base.filters.char_filter import CharFilter
from jet_bridge_base.filters.filter import EMPTY_VALUES
from jet_bridge_base.utils.queryset import get_session_engine


def get_query_func_by_name(name, column):
    if name == 'count':
        return func.count(column)
    elif name == 'sum':
        return func.sum(column)
    elif name == 'min':
        return func.min(column)
    elif name == 'max':
        return func.max(column)
    elif name == 'avg':
        return func.avg(column)

date_trunc_options = {
    'microsecond': 'microseconds',
    'millisecond': 'milliseconds',
    'second': 'minute',
    'minute': 'minute',
    'hour': 'hour',
    'day': 'day',
    'week': 'week',
    'month': 'month',
    'quarter': 'quarter',
    'year': 'year'
}

strftime_options = {
    'microseconds': '%Y-%m-%d %H:%i:%s.%f',
    # 'milliseconds': '%Y-%m-%d %H:%i:%s.%f',
    'second': '%Y-%m-%d %H:%i:%s',
    'minute': '%Y-%m-%d %H:%i:00',
    'hour': '%Y-%m-%d %H:00:00',
    'day': '%Y-%m-%d',
    # 'week': '%Y-%m-%d',
    'month': '%Y-%m-01',
    # 'quarter': '%Y-%m-%d',
    'year': '%Y-01-01'
}

dateadd_options = {
    'millisecond': 'millisecond',
    'second': 'minute',
    'minute': 'minute',
    'hour': 'hour',
    'day': 'day',
    'week': 'week',
    'month': 'month',
    'quarter': 'quarter',
    'year': 'year'
}


def get_query_lookup_func_by_name(session, lookup_type, lookup_param, column):
    try:
        if lookup_type == 'date':
            date_group = lookup_param or 'day'

            if get_session_engine(session) == 'postgresql':
                if date_group in date_trunc_options:
                    return func.date_trunc(date_trunc_options[date_group], column)
            elif get_session_engine(session) == 'mysql':
                if date_group in strftime_options:
                    return func.date_format(column, strftime_options[date_group])
            elif get_session_engine(session) == 'mssql':
                if date_group in dateadd_options:
                    interval = dateadd_options[date_group]
                    return func.dateadd(text(interval), func.datediff(text(interval), text('0'), column), text('0'))
            else:
                if date_group in strftime_options:
                    return func.strftime(strftime_options[date_group], column)
    except IndexError:
        pass

    if lookup_type:
        print('Unsupported lookup: {}'.format(lookup_type))

    return column


class ModelGroupFilter(CharFilter):

    def filter(self, qs, value):
        if value in EMPTY_VALUES:
            return qs

        x_columns = list(map(lambda x: getattr(self.model, x), value['x_columns']))
        y_column = getattr(self.model, value['y_column'])
        y_func = get_query_func_by_name(value['y_func'], y_column)

        if y_func is None:
            return qs.filter(sql.false())

        def group_name(i):
            if i == 0 and get_session_engine(qs.session) != 'mssql':
                return 'group'
            else:
                return 'group_{}'.format(i + 1)

        def map_lookup(column, i):
            lookup_name = value['x_lookups'][i] if i < len(value['x_lookups']) else None
            lookup_params = lookup_name.split('_') if lookup_name else []
            lookup_type = lookup_params[0] if len(lookup_params) >= 1 else None
            lookup_param = lookup_params[1] if len(lookup_params) >= 2 else None

            if lookup_type == 'date':
                mapper = inspect(self.model)
                column = mapper.columns.get(column.name)
                if column is not None and not isinstance(column.type, (sqltypes.DateTime, sqltypes.Date)):
                    raise ValidationError('Can\'t apply date functions to non-date field: {}'.format(column.name))

            lookup = get_query_lookup_func_by_name(qs.session, lookup_type, lookup_param, column)
            return lookup.label(group_name(i))

        x_lookups = list(map(lambda x: map_lookup(x[1], x[0]), enumerate(x_columns)))
        x_lookup_names = list(map(lambda x: x.name, x_lookups))

        whereclause = qs.whereclause
        qs = qs.session.query(*x_lookups, y_func.label('y_func'))

        if whereclause is not None:
            qs = qs.filter(whereclause)

        if get_session_engine(qs.session) == 'mssql':
            return qs.group_by(*x_lookups).order_by(*x_lookup_names)
        else:
            return qs.group_by(*x_lookup_names).order_by(*x_lookup_names)



import threading
import time
from datetime import timedelta

from graphql import GraphQLError

from jet_bridge_base.db import connection_cache
from jet_bridge_base.exceptions.permission_denied import PermissionDenied
from jet_bridge_base.logger import logger
from jet_bridge_base.permissions import HasProjectPermissions
from jet_bridge_base.responses.json import JSONResponse
from jet_bridge_base.utils.common import get_random_string
from jet_bridge_base.utils.graphql import GraphQLSchemaGenerator
from jet_bridge_base.utils.track_database import track_database_async
from jet_bridge_base.views.base.api import APIView


class GraphQLView(APIView):
    permission_classes = (HasProjectPermissions,)

    def before_dispatch_permissions_check(self, request):
        pass

    def required_project_permission(self, request):
        return {
            'permission_type': 'model',
            'permission_object': request.context.get('model'),
            'permission_actions': 'r'
        }

    def map_gql_error(self, error):
        if isinstance(error, GraphQLError):
            return error.message
        else:
            return str(error)

    def wait_schema(self, request, schema_key, wait_schema):
        if not wait_schema:
            return

        logger.info('Waiting GraphQL schema "{}"...'.format(wait_schema['id']))

        generated_condition = wait_schema['generated']
        with generated_condition:
            timeout = timedelta(minutes=10).total_seconds()
            generated_condition.wait(timeout=timeout)

        with connection_cache(request) as cache:
            cached_schema = cache.get(schema_key)
            if cached_schema and cached_schema['instance']:
                logger.info('Found GraphQL schema "{}"'.format(wait_schema['id']))
                return cached_schema['instance']
            else:
                logger.info('Not found GraphQL schema "{}"'.format(wait_schema['id']))

    def create_schema_object(self):
        new_schema_id = get_random_string(32)
        new_schema_generated = threading.Condition()
        return {'id': new_schema_id, 'instance': None, 'get_schema_time': None, 'generated': new_schema_generated}

    def create_schema(self, request, schema_key, new_schema, draft):
        try:
            logger.info('Generating GraphQL schema "{}"...'.format(new_schema['id']))

            def before_resolve(request, mapper, *args, **kwargs):
                request.context['model'] = mapper.selectable.name
                self.check_permissions(request)

            get_schema_start = time.time()
            schema = GraphQLSchemaGenerator().get_schema(request, draft, before_resolve=before_resolve)
            get_schema_end = time.time()
            get_schema_time = round(get_schema_end - get_schema_start, 3)

            with connection_cache(request) as cache:
                cached_schema = cache.get(schema_key)

                if cached_schema and cached_schema['id'] == new_schema['id']:
                    new_schema = {**new_schema, 'instance': schema, 'get_schema_time': get_schema_time}
                    cache[schema_key] = new_schema

                    logger.info('Saved GraphQL schema "{}"'.format(new_schema['id']))
                else:
                    logger.info('Ignoring GraphQL schema result "{}", existing: "{}"'.format(
                        new_schema['id'],
                        cached_schema.get('id') if cached_schema else None
                    ))

            return schema
        except Exception as e:
            with connection_cache(request) as cache:
                cached_schema = cache.get(schema_key)

                if cached_schema and cached_schema['id'] == new_schema['id']:
                    del cache[schema_key]

            raise e
        finally:
            generated_condition = new_schema['generated']
            with generated_condition:
                generated_condition.notify_all()

    def get_schema(self, request, draft):
        schema_key = 'graphql_schema_draft' if draft else 'graphql_schema'
        wait_schema = None

        with connection_cache(request) as cache:
            cached_schema = cache.get(schema_key)

            if cached_schema and cached_schema['instance']:
                return cached_schema['instance']
            elif cached_schema and not cached_schema['instance']:
                wait_schema = cached_schema
            else:
                new_schema = self.create_schema_object()
                cache[schema_key] = new_schema

        if wait_schema:
            existing_schema = self.wait_schema(request, schema_key, wait_schema)
            if existing_schema:
                return existing_schema
            else:
                with connection_cache(request) as cache:
                    new_schema = self.create_schema_object()
                    cache[schema_key] = new_schema

        return self.create_schema(request, schema_key, new_schema, draft)

    def get(self, request, *args, **kwargs):
        return self.post(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        track_database_async(request)

        draft = bool(request.get_argument('draft', False))
        validate = bool(request.data.get('validate', True))

        if 'query' not in request.data:
            return JSONResponse({})

        try:
            schema = self.get_schema(request, draft)
        except Exception as e:
            return JSONResponse({'errors': ['Failed to get table schema: {}'.format(e)]})

        query = request.data.get('query')
        context_value = {
            'request': request,
            'session': request.session
        }
        result = schema.execute(
            query,
            variables={},
            context_value=context_value,
            validate=validate
        )

        if result.errors is not None and len(result.errors):
            error = result.errors[0]
            if hasattr(error, 'original_error'):
                error = error.original_error
            if isinstance(error, PermissionDenied):
                raise error
            return JSONResponse({'errors': map(lambda x: self.map_gql_error(x), result.errors)})

        return JSONResponse({
            'data': result.data
        })
